/* Copyright (C) 2000-2004  Thomas Bopp, Thorsten Hampel, Ludger Merkens
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * $Id: auth.pike,v 1.2 2005/10/26 11:32:32 astra Exp $
 */

inherit "/kernel/module";

#include <classes.h>
#include <database.h>
#include <events.h>
#include <macros.h>

static object usermod = get_module("users");
static object ldap = get_module("ldap");

/**
 * Find a user object if the submitted password is sent correctly.
 *  
 * @param string name - the user name
 * @param string pass - the users password
 * @return the user object or null, throws an error if the password is wrong
 */
object authenticate(string name, string pass)
{
  object user = _Persistence->lookup_user( name );
  if ( objectp(user) ) {
    if ( user->get_activation() )
      steam_user_error("User "+name+" was not activated.");
    if(!ldap)
      ldap = get_module("ldap");
    if ( objectp(ldap) && ldap->authorize_ldap(user, pass) )
      return user;
    if ( user->check_user_password(pass) )
      return user;
  }
  else
    MESSAGE( "User authorization failed for \"%s\".", name );
  return 0;
}

string get_ticket(object user)
{
  // try to get ticket from user and possible store inside ldap
}

string get_identifier() { return "auth"; }
