/* Copyright (C) 2000-2005  Thomas Bopp, Thorsten Hampel, Ludger Merkens
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
//inherit "/kernel/module";
inherit "/kernel/secure_mapping.pike";
inherit "/kernel/persistence.pike";

#include <macros.h>
#include <classes.h>
#include <database.h>
#include <exception.h>
#include <attributes.h>
#include <configure.h>

string get_identifier() { return "persistence:ldap"; }
object get_environment() { return 0; }

static object      oLDAP;
static mapping    config;

static int persistence_id;
static int oid_length = 4;  // nr of bytes for object id

static void init_module()
{
  persistence_id = _Persistence->register(this_object());
  config = read_config(Stdio.read_file(CONFIG_DIR+"/ldap.txt"), "ldap");
  connect_ldap();
}

int get_persistence_id ()
{
  return persistence_id;
}

private static void connect_ldap()
{
  if ( !stringp(config["server"]) )
    return;

  mixed err = catch {
    oLDAP = Protocols.LDAP.client(config["server"]);
    
    if ( stringp(config["user"]) && sizeof(config["user"])>0 )
      oLDAP->bind("cn="+config["user"] + 
		  (stringp(config->root_dn) &&
		   strlen(config->root_dn)>0?","+config->root_dn:""),
		  config["password"]);
     
    if ( oLDAP->error_number() > 0 )
      FATAL("Failed to bind ldap: " + oLDAP->error_string());	
    oLDAP->set_scope(2);  // subtree scope
    oLDAP->set_basedn(config["base_dc"]);
  };
  if ( err != 0 )
    FATAL("Failed to build connection to LDAP: %s\n%O", config->server, err);
  if ( objectp(oLDAP) && oLDAP->error_number() > 0 )
    FATAL("Failed to bind ldap: " + oLDAP->error_string());	
}

mixed lookup_user(string identifier)
{
  if ( !objectp(oLDAP) )
    return 0;

  mixed data = get_module("ldap")->fetch_user(identifier);

  if ( !mappingp(data) ) return 0;

  mixed exc = catch {

  mapping result = ([ "ldap:dn" : data["dn"] ]);
  if(data[config->passwordAttr])
    result->UserPassword = data[config->passwordAttr];
  if(data[config->emailAttr])
    result[USER_EMAIL] = data[config->emailAttr];
  if(data[config->fullnameAttr])
    result[USER_FULLNAME] = data[config->fullnameAttr];
  if(data[config->nameAttr])
    result[USER_FIRSTNAME] = data[config->nameAttr];
  if(data[config->descriptionAttr])
    result[OBJ_DESC] = data[config->descriptionAttr];

  if ( stringp(config->groupAttr) && sizeof(config->groupAttr)>0 )
  {
      // primary group:
      if ( stringp(config->groupId) && stringp(data[config->groupId]) ) {
	  string searchstr = "(&(" + config->groupId + "=" + data[config->groupId] + ")"
	      + "(objectClass=" + config->groupClass + "))";
	  mixed primary_group = get_module("ldap")->fetch( "", searchstr );
	  if ( mappingp(primary_group) && stringp(config->groupAttr)
	       && stringp(primary_group[config->groupAttr]) )
	      result["ActiveGroup"] = primary_group[config->groupAttr];
      }
      // groups:
      if ( stringp(config->memberAttr) ) {
	  array groups = ({ "sTeam" });
	  string searchstr = "(&(" + config->memberAttr + "=" + identifier + ")"
	      + "(objectClass=" + config->groupClass + "))";
	  mixed ldap_groups = get_module("ldap")->fetch( "", searchstr );
	  if ( mappingp(ldap_groups) )  // single group, make an array out of it:
	      ldap_groups = ({ ldap_groups });

	  foreach ( ldap_groups, mixed ldap_group ) {
	      if ( !mappingp(ldap_group) ) {
		  werror( "ldappersistence: invalid ldap group: %O\n", ldap_group );
		  continue;
	      }
	      if ( stringp(ldap_group[config->groupAttr]) )
		  groups += ({ ldap_group[config->groupAttr] });
	      else
		  werror( "ldappersistence: invalid ldap group name: %O\n", ldap_group[config->groupAttr] );
	  }
	  result["Groups"] = groups;
      }
  }

  return result;

  };
  if ( exc != 0 ) werror( "LDAP: lookup_user(\"%s\") : %O\n", identifier, exc );
}

mixed lookup_group(string identifier)
{
  if ( !objectp(oLDAP) )
    return 0;

  if ( !stringp(config->groupAttr) || sizeof(config->groupAttr)<1 )
    return 0;

  mixed data = get_module("ldap")->fetch_group(identifier);
  if ( !mappingp(data) ) return 0;

  mixed exc = catch {

  mapping result = ([
    "ldap:dn" : data["dn"],
    "OBJ_DESC" : data[config->descriptionAttr],
  ]);
  // users:
  if ( stringp(config->memberAttr) ) {
    mixed ldap_users = data[config->memberAttr];
    if ( stringp(ldap_users) )  // single user, make an array out of it:
      ldap_users = ({ ldap_users });
    result["Users"] = ldap_users;
  }

  return result;

  };
  if ( exc != 0 ) werror( "LDAP: lookup_user(\"%s\") : %O\n", identifier, exc );
}

void update_user ( object user )
{
  //TODO: implement update_user(obj) ?
  //werror("* server/modules/ldappersistence.pike : update_user() not implemented yet\n" );
}

void update_group ( object group )
{
  //TODO: implement update_group(obj) ?
  //werror("* server/modules/ldappersistence.pike : update_group() not implemented yet\n" );
}
